SET ANSI_NULLS, ANSI_PADDING, ANSI_WARNINGS, ARITHABORT, CONCAT_NULL_YIELDS_NULL, QUOTED_IDENTIFIER ON;
SET NUMERIC_ROUNDABORT OFF;
GO

IF NOT EXISTS (
        SELECT NULL
        FROM sys.indexes
        WHERE name = 'IX_Versioned_Locks_Filtered'
            AND object_id = OBJECT_ID('VersionedFields')
        )
BEGIN
    PRINT N'Creating [VersionedFields].IX_Versioned_Locks_Filtered index...'
    CREATE NONCLUSTERED INDEX IX_Versioned_Locks_Filtered ON [dbo].[VersionedFields] ([ItemID]) INCLUDE (
        [Value]
        ,[Language]
        ,[Version]
        )
    WHERE [FieldId] = '{001DD393-96C5-490B-924A-B0F25CD9EFD8}'
END


/* Conditionally update relevant script for FORM */
IF (EXISTS (SELECT * 
                 FROM INFORMATION_SCHEMA.TABLES 
                 WHERE TABLE_SCHEMA = 'dbo' 
                 AND  TABLE_NAME = 'SharedFields') AND EXISTS (SELECT * FROM SharedFields WHERE VALUE LIKE '%[A-Z0-9][-][0-2]%[.][A-Z0-9]%' AND FieldId = 'FCE03B41-4CBB-4F7F-920F-FF04F12FA896'))
BEGIN
	BEGIN TRAN
	UPDATE d
	SET value = Replace(d.value,SUBSTRING(v.value,front.pos,txt.ln) +'.js','jquery-3.4.1.min.js') +'|form.conditions.js'
	FROM SharedFields as d
	INNER JOIN SharedFields v ON v.id = d.id AND v.FieldId = 'FCE03B41-4CBB-4F7F-920F-FF04F12FA896' 
	CROSS APPLY (VALUES(CHARINDEX('jquery-',v.value))) AS front(pos)
	CROSS APPLY (VALUES(CHARINDEX('.js',v.value,front.pos)-front.pos)) AS txt(ln)
	WHERE v.value like '%[A-Z0-9][-][0-9]%[.][A-Z0-9]%'   
	AND  d.value like '%[A-Z0-9][-][0-9]%[.][A-Z0-9]%'   
	AND d.FieldId = 'FCE03B41-4CBB-4F7F-920F-FF04F12FA896' 
	COMMIT TRAN
END

GO

/*
This script to perform update on the EXM existing rootmanagers. Updates include 
1. Remove obsoleted templates
2. Add new templates 
*/

CREATE FUNCTION [dbo].[EXM_STRING_SPLIT] 
(
    @stringList nvarchar(max), 
    @separator nvarchar(1)
)
RETURNS 
@result TABLE ([value] nvarchar(max) )
AS
BEGIN

	DECLARE @x XML
	SELECT @X = CAST('<A>' + REPLACE(@stringList, @separator, '</A><A>') + '</A>' AS XML)

	INSERT INTO @Result
	SELECT t.value('.', 'nvarchar(max)') as inVal
	FROM @X.nodes('/A') AS x(t)
		RETURN 
END

GO 

IF (EXISTS (SELECT * 
                 FROM INFORMATION_SCHEMA.TABLES 
                 WHERE TABLE_SCHEMA = 'dbo' 
                 AND  TABLE_NAME = 'SharedFields'))
BEGIN
Print 'Updating SharedFields for InsertOptions for EXM Items.'

Declare @RootListId varchar(40) = '684CAA9A-FBD3-4FFB-891E-042E56B4E09C'
Declare @RootListFieldId varchar(40) = '9E4027B2-0360-4B48-B665-FC51E55FC766'
Declare @RegularMessageTemplateId varchar(40) = '54E7880D-2621-49EA-A417-982FCD372903'
Declare @AutomatedMessageTemplateId varchar(40) = '9795C562-98D9-4A9E-BC9C-C6B7A079FDF4'
Declare @InsertOptionFieldId varchar(40) = '1172F251-DAD4-4EFB-A329-0C63500E4F1E'
Declare @NewTemplateList nvarchar(500) = '{8C18055A-EB1E-4358-BC12-26B4E2568910}|{FD58C821-8FAA-4C51-B368-5FC25CA6AAE0}|{89F09F47-6ACA-41A9-ACC6-2E33A0334C0C}|{56246CDE-5320-4C15-87E9-28DF4A0133D1}|{F868C724-50D0-4347-9C9B-BABB8DB705FE}|{A0F3FEFA-D89D-4F5F-9406-E9D1FA7143DF}|{90A769C7-47DA-4042-9F49-413BABC66504}|{7CD7EA23-F1DF-4A1E-A0B3-0418FCF7BCC7}|{182B4D7D-803A-43C6-B3E0-F1E61A606153}|{0E6E9CE3-29D6-4564-B13B-2327F77543F5}|{2FCD032A-669B-4750-B617-2A38F95E0EA7}'
Declare @OldTemplateList nvarchar(500) = '{6FE51EB4-1D30-4E6B-8BA0-0EBB1405D283}|{9E170B31-F8AA-4E1F-B605-2FC0BD4FE120}|{ECF0A9C8-FE56-45D8-B750-6BDE3A34A532}|{1AFE38A7-9461-4278-ADAF-D807F27F36E4}'

DECLARE @UserTable TABLE ([id] uniqueidentifier, [Value] nvarchar(MAX), [Index] INT)
DECLARE @NewTemplatesTable TABLE ( [Value] nvarchar(MAX))
DECLARE @OldTemplatesTable TABLE ( [Value] nvarchar(MAX))

Insert into @NewTemplatesTable Select value from EXM_STRING_SPLIT(@NewTemplateList,'|')
Insert into @OldTemplatesTable Select value from EXM_STRING_SPLIT(@OldTemplateList,'|')

Insert into @UserTable
SELECT [Id],[Value],ROW_NUMBER() OVER (order by Id) as [Index]

  FROM [dbo].[SharedFields]
  where  FieldId =  @InsertOptionFieldId
  AND ItemId in(
	SELECT  [ID] FROM [dbo].[Items]
	where ID IN (	
		SELECT [Descendant]
		FROM [dbo].[Descendants]
		where Ancestor in (	SELECT value as ID 
			FROM EXM_STRING_SPLIT(
				(SELECT [Value]
				FROM [dbo].[SharedFields]
				where ItemId = @RootListId and FieldId = @RootListFieldId)
			, '|')
			WHERE RTRIM(value) <> '' ) ) and (TemplateID = @RegularMessageTemplateId OR TemplateID = @AutomatedMessageTemplateId))	

DECLARE @count INT = (select count(value) from @UserTable);
DECLARE @index INT = 0;

BEGIN TRAN

WHILE @count > @index
	BEGIN

	   SET @index = @index + 1;

		Declare @ValueField nvarchar(max) = ''
		DECLARE @FinalValue nvarchar(max) = ''
		DECLARE @CurrentId uniqueidentifier = null

		Select @CurrentId = [Id] from @UserTable where [Index] = @index
		Select @ValueField = [Value] from @UserTable where [Index] = @index
		Select @FinalValue = COALESCE(@FinalValue + '|' + [value], [value]) 
			From (select value from EXM_STRING_SPLIT(@ValueField,'|') union Select value from  @NewTemplatesTable EXCEPT Select [value] from @OldTemplatesTable ) as FinalValueTable;
			Select @FinalValue = substring(@FinalValue,len('|')+1,len(@FinalValue))
		update [dbo].[SharedFields] set [Value]=@FinalValue where Id = @CurrentId
	
	END

COMMIT TRAN
END
ELSE
BEGIN
	Print 'Failed to update EXM. Table does not exists.'
END

GO

IF EXISTS (SELECT *
           FROM   sys.objects
           WHERE  object_id = OBJECT_ID(N'[dbo].[EXM_STRING_SPLIT]')
                  AND type IN ( N'FN', N'IF', N'TF', N'FS', N'FT' ))
  DROP FUNCTION [dbo].[EXM_STRING_SPLIT]

GO

PRINT N'Update complete.';