<%@ Page Language="c#" Inherits="System.Web.UI.Page" CodePage="65001" Debug="true"%>
<%@ OutputCache Location="None" VaryByParam="none" %>
<%@ Import Namespace="System.Globalization" %>
<%@ Import Namespace="Sitecore.DependencyInjection" %>
<%@ Import Namespace="Microsoft.Extensions.DependencyInjection" %>
<%@ Import Namespace="Sitecore.Xdb.ReferenceData.Core" %>
<%@ Import Namespace="Sitecore.Marketing.Operations.Xdb.ReferenceData.Model.Definitions" %>
<%@ Import Namespace="Sitecore.Marketing.Operations.Xdb.ReferenceData.Model.Definitions.Segments" %>
<%@ Import Namespace="Sitecore.Marketing.xMgmt.Definitions.Segments.Data.ItemDb" %>
<%@ Import Namespace="Sitecore.Xdb.ReferenceData.Core.Collections" %>
<%@ Import Namespace="Sitecore.Marketing.Definitions" %>
<%@ Import Namespace="Sitecore.Marketing.Definitions.Segments.Data" %>
<%@ Import Namespace="Sitecore.Xdb.ReferenceData.Core.Converter" %>

<script runat="server">
  private IMonikerConverter<Guid> monikerConverter = ServiceLocator.ServiceProvider.GetService<IMonikerConverter<Guid>>();
  private IReferenceDataClient referenceDataClient = ServiceLocator.ServiceProvider.GetService<IReferenceDataClient>();
  private SegmentDefinitionItemRepository segmentDefinitionItemRepository = ActivatorUtilities.CreateInstance<SegmentDefinitionItemRepository>(ServiceLocator.ServiceProvider);

  protected void Run_Click(object sender, EventArgs e)
  {
    var segmentsToDeactivate = GetSegmentsToDeactivate();
    foreach (var segmentToDeactivate in segmentsToDeactivate)
    {
      segmentToDeactivate.IsActive = false;
    }

    if (segmentsToDeactivate.Any())
    {
      referenceDataClient.Save(segmentsToDeactivate);

      RenderSummary(segmentsToDeactivate);
      this.Summary.Text = @"Next definitions were deactivated in Reference Data: <br><br>" + this.Summary.Text;
    }
    else
    {
      this.Summary.Text = "Nothing to deactivate.";
    }
    this.Run.Text = "Done, please see summary below.";
    this.TestRun.Text = "Test Run - Get affected segments.";
  }

  protected void TestRun_Click(object sender, EventArgs e)
  {
    var segmentsToDeactivate = GetSegmentsToDeactivate();
    RenderSummary(segmentsToDeactivate);
    if (segmentsToDeactivate.Any())
    {
      this.Summary.Text = @"Next definitions will be deactivated in Reference Data: <br><br>" + this.Summary.Text;
    }
    this.TestRun.Text = "Done, please see summary below.";
    this.Run.Text = "Deactivate segments.";
  }

  private void RenderSummary(DefinitionCollection<SegmentCommonData, SegmentCultureData> segmentsToDeactivate)
  {
    if (segmentsToDeactivate.Any())
    {
      this.Summary.Text = "Count=" + segmentsToDeactivate.Count + "<br>" + string.Join("<br>", segmentsToDeactivate.Select(d => "Alias=" + d.CommonData.Alias + " Id=" + d.Key.Moniker + " Version=" + d.Key.Version));
    }
    else
    {
      this.Summary.Text = "Nothing to deactivate.";
    }
  }

  private DefinitionCollection<SegmentCommonData, SegmentCultureData> GetSegmentsToDeactivate()
  {
    var segmentDefinitionTypeKey = referenceDataClient.EnsureDefinitionType(DefinitionTypeNames.Segment);
    var segmentDefinitions = referenceDataClient.GetDefinitionsByType<SegmentCommonData, SegmentCultureData>(
      segmentDefinitionTypeKey,
      false,
      1,
      int.MaxValue).Definitions.Where(d => d.IsActive).ToList();

    var segmentItems = segmentDefinitionItemRepository.GetAll(CultureInfo.InvariantCulture, new RetrievalParameters<SegmentDefinitionRecord, string>(pageSize: int.MaxValue), true);

    var segmentsToDeactivate = new DefinitionCollection<SegmentCommonData, SegmentCultureData>();
    foreach (var segmentDefinition in segmentDefinitions)
    {
      var segmentId = monikerConverter.Convert(segmentDefinition.Key.Moniker);
      if (!segmentItems.Any(si => si.Data.Id == segmentId))
      {
        segmentsToDeactivate.Add(segmentDefinition);
      }
    }

    return segmentsToDeactivate;
  }
</script>

<html>
  <head>
    <title>Deactivate segment if not exists on CM</title>
    <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
    <meta name="CODE_LANGUAGE" content="C#" />
    <meta name="vs_defaultClientScript" content="JavaScript" />
    <meta name="vs_targetSchema" content="http://schemas.microsoft.com/intellisense/ie5" />
  </head>
  <body>
  <script>
    function OnTestRunClick() {
      this.disabled = true;
      this.value = 'Getting segments. Please wait...';
      return true;
    }
    function OnRunDeactivationClick() {
      if (confirm('Are you sure?')) {
        this.disabled = true;
        this.value = 'Deactivating. Please wait...';
        return true;
      } else {
        return false;
      }
    }
  </script>
  <form method="post" runat="server">
    <h2>Reads all active segments from ReferenceData and deactivates those not existing under /sitecore/system/Marketing Control Panel/Segments</h2>
    <p>
      Note: This page deactivates segment definition in ReferenceData if segment item is not found under Marketing Control Panel/Segments,
      any segment item not from Marketing Control Panel/Segments will be treated as 'not found' hence deactivated in ReferenceData.
    </p>
    <asp:Button ID="TestRun" runat="server" Text="Test Run - Get affected segments" OnClick="TestRun_Click" OnClientClick="return OnTestRunClick();" />
    <asp:Button ID="Run" runat="server" Style="margin-left: 30px" Text="Deactivate segments" OnClick="Run_Click" OnClientClick="return OnRunDeactivationClick();" />
    <p>
    <asp:Label ID="Summary" Width="1000px" Height="500px" runat="server" />
  </form>
  </body>
</html>
