﻿<%@ Page Language="c#" Inherits="Sitecore.sitecore.admin.AdminPage" CodePage="65001" %>
<%@ OutputCache Location="None" VaryByParam="none" %>

<%@ Import Namespace="System.Data" %>
<%@ Import Namespace="System.Data.SqlClient" %>
<%@ Import Namespace="System.IO" %>
<%@ Import Namespace="System.Reflection" %>
<%@ Import Namespace="Sitecore.Globalization" %>
<%@ Import Namespace="Sitecore.Data" %>

<%@ Import Namespace="Sitecore.Configuration" %>
<%@ Import Namespace="Sitecore.Diagnostics" %>
<%@ Import Namespace="Sitecore.Data.Managers" %>
<%@ Import Namespace="Sitecore.Data.Items" %>

<!DOCTYPE html>

<script runat="server">

    private readonly Func<object, ID> _parseIdFunc = Sitecore.Data.ID.Parse;

    private const string FileMask = "RestoredTaxonomies_{0}.dat";

    private const string DatabaseName = "master";

    private string _dumpFolder;

    protected override void OnInit(EventArgs arguments)
    {
        CheckSecurity(true);
        base.OnInit(arguments);
    }

    protected void Page_Load(object sender, EventArgs e)
    {
        _dumpFolder = Server.MapPath("~/App_Data");
    }

    protected void Page_PreRender(object sender, EventArgs e)
    {
        var files = new string[0];
        if (Directory.Exists(_dumpFolder))
        {
            files = Directory.EnumerateFiles(_dumpFolder, string.Format(FileMask, "*")).ToArray();
        }

        FileRepeater.DataSource = files;
        FileRepeater.DataBind();
    }

    protected void btnRestore_Click(object sender, EventArgs eventArgs)
    {
        spanRestored.Visible = true;
        spanRestored.InnerHtml = string.Format("Restored {0} items.",
            RestoreTaxonomies(Path.Combine(_dumpFolder, string.Format(FileMask, DateTime.UtcNow.ToString("yyyy_MM_dd_HH_mm_ss")))));
    }

    protected void btnCleanup_Click(object sender, CommandEventArgs e)
    {
        spanRestored.Visible = true;
        spanRestored.InnerHtml = string.Format("Removed {0} items.",
            new CommandDump(e.CommandArgument.ToString()).LoadAndUndo(DatabaseName, true));
    }

    private int RestoreTaxonomies(string fileName)
    {
        var connectionString = Settings.GetConnectionString("reporting");

        const string entitiesCmd = "SELECT t.Id, t.ParentId, t.Type, f.FieldValue FROM Taxonomy_TaxonEntity AS t" +
                                   " LEFT OUTER JOIN Taxonomy_TaxonEntityFieldValue as f on f.TaxonId = t.Id AND f.FieldId = @ItemNameFieldId AND f.LanguageCode = 'en'" +
                                   "  ORDER BY LEN (t.Uri)"; // order by is required to create top level taxonomy items first.
        // Otherwise additional logic for restoring all taxon parents is required.

        var masterDb = Factory.GetDatabase(DatabaseName);
        var commands = new Dictionary<ID, RestoreCommand>();
        using (var connection = new SqlConnection(connectionString))
        {
            using (var command = new SqlCommand(entitiesCmd, connection))
            {
                connection.Open();
                command.Parameters.Add("@ItemNameFieldId", SqlDbType.UniqueIdentifier);
                command.Parameters["@ItemNameFieldId"].Value = Guid.Parse(Sitecore.Marketing.Taxonomy.WellKnownIdentifiers.Entity.TaxonFields.NameFieldId.ToString());
                using (var reader = command.ExecuteReader())
                {
                    while (reader.Read())
                    {
                        var itemId = _parseIdFunc(reader["Id"]);
                        commands.Add(itemId,
                            new RestoreItemCommand(
                                itemId,
                                reader["FieldValue"].ToString(),
                                _parseIdFunc(reader["ParentId"]),
                                _parseIdFunc(reader["Type"]),
                                masterDb));
                    }
                }
            }
        }

        var dump = new CommandDump(fileName);

        foreach (var command in commands)
        {
            if (dump.ExecuteAndAdd(command.Value))
            {
                RestoreLanguages(dump, command.Key, masterDb);
            }
        }

        return dump.Flush();
    }

    private void RestoreLanguages(CommandDump dump, ID itemId, Database database)
    {
        var connectionString = Settings.GetConnectionString("reporting");
        const string entityFields = "SELECT f.FieldId, f.FieldValue, f.LanguageCode FROM Taxonomy_TaxonEntity as t" +
                                    " Left outer join Taxonomy_TaxonEntityFieldValue as f on f.TaxonId = t.Id" +
                                    " WHERE f.TaxonId = @Id And f.FieldId <> @ItemNameFieldId";

        var languages = new Dictionary<string, Dictionary<ID, string>>();
        using (var connection = new SqlConnection(connectionString))
        using (var command = new SqlCommand(entityFields, connection))
        {
            command.Parameters.Add("@Id", SqlDbType.UniqueIdentifier);
            command.Parameters["@Id"].Value = itemId.Guid;

            command.Parameters.Add("@ItemNameFieldId", SqlDbType.UniqueIdentifier);
            command.Parameters["@ItemNameFieldId"].Value = Guid.Parse(Sitecore.Marketing.Taxonomy.WellKnownIdentifiers.Entity.TaxonFields.NameFieldId.ToString());
            connection.Open();
            using (var reader = command.ExecuteReader())
            {
                while (reader.Read())
                {
                    var language = reader["LanguageCode"].ToString();
                    Dictionary<ID, string> fields;
                    if (!languages.TryGetValue(language, out fields))
                    {
                        fields = languages[language] = new Dictionary<ID, string>();
                    }

                    fields.Add(_parseIdFunc(reader["FieldId"]), reader["FieldValue"].ToString());
                }
            }
        }

        foreach (var language in languages)
        {
            dump.ExecuteAndAdd(new RestoreLanguageCommand(itemId, language.Key, language.Value, database));
        }
    }

    private class CommandDump
    {
        private readonly string _file;
        private const char TypeSeparator = ':';

        private readonly List<string> _list = new List<string>();

        public CommandDump(string file)
        {
            _file = file;
        }

        public bool ExecuteAndAdd(RestoreCommand command)
        {
            var result = command.Execute();
            if (result)
            {
                _list.Add(string.Format("{0}:{1}", command.GetType().Name, command));
            }

            return result;
        }

        public int Flush()
        {
            var directoryName = Path.GetDirectoryName(_file);

            if (!Directory.Exists(directoryName))
            {
                Directory.CreateDirectory(directoryName);
            }

            File.WriteAllLines(_file, _list);
            return _list.Count;
        }

        public int LoadAndUndo(string databaseName, bool removeFile)
        {
            var lines = File.ReadAllLines(_file);
            foreach (var line in lines)
            {
                var strings = line.Split(TypeSeparator);
                var type = Assembly.GetExecutingAssembly().GetTypes().Single(x => x.Name == strings[0]);

                var @params = RestoreCommand.ParseParams(strings[1]);
                @params["database"] = databaseName;

                var command = (RestoreCommand)Activator.CreateInstance(type, @params);
                command.Revert();
            }

            if (removeFile)
            {
                File.Delete(_file);
            }

            return lines.Length;
        }
    }

    private abstract class RestoreCommand
    {
        private const string ParamSeparator = ";";
        private const char ParamValueSeparator = '=';

        public abstract bool Execute();
        public abstract void Revert();

        public abstract Dictionary<string, string> GetParams();

        public static Dictionary<string, string> ParseParams(string str)
        {
            return str.Split(new[] {ParamSeparator}, StringSplitOptions.None).ToDictionary(x => x.Split(ParamValueSeparator)[0], x => x.Split(ParamValueSeparator)[1]);
        }

        public override string ToString()
        {
            return DictionaryToString(GetParams(), ParamSeparator, ParamValueSeparator.ToString());
        }
    }

    private class RestoreLanguageCommand : RestoreCommand
    {
        private readonly ID _itemId;
        private readonly string _language;
        private readonly Dictionary<ID, string> _fields;
        private readonly Database _db;

        public RestoreLanguageCommand(Dictionary<string, string> args)
            : this(Sitecore.Data.ID.Parse(args["itemId"]), args["language"], new Dictionary<ID, string>(), Factory.GetDatabase(args["database"]))
        {
        }

        public RestoreLanguageCommand(ID itemId, string language, Dictionary<ID, string> fields, Database db)
        {
            _itemId = itemId;
            _language = language;
            _fields = fields;
            _db = db;
        }

        public override bool Execute()
        {
            var item = _db.GetItem(_itemId, Language.Parse(_language));
            if (item.Versions.Count != 0)
            {
                return false;
            }

            item = item.Versions.AddVersion();

            using (new EditContext(item))
            {
                foreach (var field in _fields)
                {
                    var itemField = item.Fields[field.Key];
                    if (itemField != null)
                    {
                        itemField.SetValue(field.Value, true);
                    }
                }
            }

            return true;
        }

        public override void Revert()
        {
            var item = _db.GetItem(_itemId, Language.Parse(_language));

            if (item == null || item.Versions.Count == 0)
            {
                return;
            }

            item.Versions.RemoveVersion();
        }

        public override Dictionary<string, string> GetParams()
        {
            return new Dictionary<string, string>
            {
                {"itemId", _itemId.ToString()},
                {"language", _language},
            };
        }
    }

    private class RestoreItemCommand : RestoreCommand
    {
        private readonly ID _itemId;
        private readonly string _itemName;
        private readonly ID _parentId;
        private readonly ID _templateId;
        private readonly Database _db;

        public RestoreItemCommand(Dictionary<string, string> args)
            : this(Sitecore.Data.ID.Parse(args["itemId"]), args["itemName"], Sitecore.Data.ID.Parse(args["parentId"]), Sitecore.Data.ID.Parse(args["templateId"]), Factory.GetDatabase(args["database"]))
        {
        }

        public RestoreItemCommand(ID itemId, string itemName, ID parentId, ID templateId, Database db)
        {
            _itemId = itemId;
            _itemName = itemName;
            _parentId = parentId;
            _templateId = templateId;
            _db = db;
        }

        public override bool Execute()
        {
            if (_db.GetItem(_itemId) != null)
            {
                return false;
            }

            var parent = _db.GetItem(_parentId);
            Assert.IsNotNull(parent, "Could not find parent item '{0}' of item '{1}'!", _parentId, _itemId);

            ItemManager.CreateItem(_itemName, parent, _templateId, _itemId);
            return true;
        }

        public override void Revert()
        {
            var item = _db.GetItem(_itemId);
            if (item != null)
            {
                item.Delete();
            }
        }

        public override Dictionary<string, string> GetParams()
        {
            return new Dictionary<string, string>
            {
                { "itemId", _itemId.ToString() },
                { "itemName", _itemName },
                { "parentId", _parentId.ToString() },
                { "templateId", _templateId.ToString() }
            };
        }
    }

    private static string DictionaryToString(Dictionary<string, string> dict, string elementSeparator, string keyValueSeparator)
    {
        return string.Join(elementSeparator, dict.Select(x => x.Key + keyValueSeparator + x.Value));
    }

</script>
<style type="text/css">
    body {
        padding: 5px;
    }
    .note {
            background-color: lightgray;
    }
    #report {
        padding: 20px 0;
    }
</style>

<html lang="en">
<head>
    <title>Restore Deleted Marketing Taxonomies</title>
    <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
    <meta name="CODE_LANGUAGE" content="C#" />
    <link href="/default.css" rel="stylesheet" />
    <link rel="shortcut icon" href="/sitecore/images/favicon.ico" />
</head>
<body>
  <div id="MainPanel">
    <h1>Restore Deleted Marketing Taxonomies</h1>
    <p class="note">
        In Sitecore 9.0, marketing taxonomies are no longer deployed to the
        reporting database. Instead, they are deployed to the Reference Data
        service. As part of upgrade to 9.0 it is necessary to copy all
        marketing taxonomies from the reporting database to the Reference Data
        Service.
        <br/><br/>
        The easiest way to deploy all marketing taxonomies to the Reference
        Data service is to use the Deploy marketing taxonomies dialog (Control
        Panel -> Analytics -> Deploy marketing definitions) after Sitecore is
        upgraded to 9.0. This dialog gets all marketing taxonomies from the
        content database and deploys them to the Reference Data service.
        <br/><br/>
        However, there may be situations when a marketing taxonomy exists in
        the reporting database but the corresponding marketing taxonomy item
        no longer exists in the content database. This can happen when a
        marketing taxonomy item is created in the content database and
        deployed to the reporting database, after which the original item is
        deleted.
        <br /><br />
        In previous versions of Sitecore, taxonomies were maintained in the
        reporting database even when corresponding items were removed from the
        content database. This was done by manually copying tables before
        rebuilding the reporting database, and was required to ensure that the
        rebuild did not fail when old interactions with references to deleted
        taxonomies were processed.
        <br/><br/>
        This page gets marketing taxonomies from the reporting database and
        restores corresponding items/versions if they no longer exist in the
        content database. These restored items will be automatically updated
        during the installation of upgrade package. Once upgrade package is
        installed, all marketing taxonomies have to be deployed to the
        Reference Data service with the help of Deploy marketing definitions
        dialog.
        <br /><br />
        Once that is done, restored items/versions can be safely removed by
        using this page as it keeps track of restored items/versions. If you
        choose not to delete the restored items, make sure your marketing team
        is aware that previously deleted items were restored as part of the
        upgrade process.
        <br/><br/>
        <b>Important note:</b> This page requires direct access to the
        <code>reporting</code> database. Ensure that the current instance has
        appropriate connection string in
        <code>Website\App_Config\ConnectionStrings.config</code> file and the
        database is accessible.
    </p>
    <form method="post" runat="server" id="mainform">
            <div>
                <asp:Button
                    OnClick="btnRestore_Click"
                    runat="server"
                    Text="Restore deleted taxonomy items"
                    OnClientClick="this.disabled = true; this.value = 'Restoring. Please wait...';"
                    UseSubmitBehavior="false"/>
                <div id="report">
                    <span runat="server" id="spanRestored" Visible="False"></span>
                </div>
            </div>
            <asp:Repeater runat="server" ID="FileRepeater">
                <ItemTemplate>
                    <div>
                        <%# Container.DataItem%>
                        <asp:Button
                            CommandArgument='<%# Container.DataItem %>'
                            OnCommand="btnCleanup_Click"
                            runat="server"
                            Text="Remove restored taxonomy items"
                            OnClientClick="this.disabled = true; this.value = 'Removing. Please wait...';"
                            UseSubmitBehavior="false"/>
                    </div>
                </ItemTemplate>
            </asp:Repeater>
    </form>
  </div>
</body>
</html>
